#include "global.hpp"
#include <signal.h>

uint32_t bytes;
UCHAR data[64* 2];
FT_HANDLE handle;
ULONG written;
static int callback;
uint8_t arg_opt;
char *filename;
uint8_t fuse_val[3];

[[noreturn]] void exit_app(int exitcode) {
	FT_Close(handle);
	exit(1);
}

[[noreturn]] void signal_handler(int signum) {
	exit_bootmode();
	exit_app(0);
}

int main(int argc, char **argv){
	signal(SIGINT, signal_handler);
	parse_argument(argc, argv);

	if( arg_opt & ARG_HELP ) {
		printf("Usage: %s <opt>\n", argv[0]);
		exit(1);
	}

	if( FT_Open(0, &handle) != FT_OK ) {
		// FIXME: Connection Failure repeats.
		printf("Failed to connect FTDI device.\n");
		exit(1);
	}
	FT_SetBaudRate(handle, 115200);
	// FT_SetBaudRate(handle, 1200);
	FT_SetTimeouts(handle, 1000, 1000);
	FT_SetBitMode(handle, 0x00, FT_BITMODE_RESET);
	FT_SetBitMode(handle, 0xFF, FT_BITMODE_SYNC_BITBANG);

	enter_bootmode();
	polling_while_busy();
	// プログラミング許可
	if( write_target(BEGIN_PROGRAMMING) != FT_OK ) {
		printf("Cannot send bytes to the target.\n");;
		exit_app(1);
	}
	read_target();
	callback = expandFromArray(data);
	if( (callback & 0x0000FF00) != (0x53 << 8) ) {
		printf("Failed to start programming\n");
		exit_app(1);
	}

	// デバイス情報・ヒューズビットをask
	ask_devinfo();
	printf("\n");
	printf("Data at 0x0000: 0x%" PRIX32 "\n", readAt(0x00));

	if( arg_opt & ARG_READ ) {
		uint16_t pages = 0;
		for(int i=0; i<256*64; i++){
			if( i % 64 == 0 ) {
				printf("\n#0x%04" PRIX32 ": ", pages);
			}
			printf("%04" PRIX16 " ", readAt(i));
			if( i % 64 == 0 ) {
				pages++;
			}
		}
	}
	if( arg_opt & ARG_LFUSE ) {
		printf("Change LFUSE to 0x%X. OK? [y/N] > ", fuse_val[FUSE_L]);
		char ans = getchar();
		if( ans == 'y' ) {
			write_lfuse(fuse_val[FUSE_L]);
			read_target();
			polling_while_busy();
		}else{
			printf("Operation cancelled.\n");
			exit_bootmode();
			exit_app(1);
		}
	}
	if( arg_opt & ARG_HFUSE ) {
		printf("Change HFUSE to 0x%X. OK? [y/N] > ", fuse_val[FUSE_H]);
		char ans = getchar();
		if( ans == 'y' ) {
			write_hfuse(fuse_val[FUSE_H]);
			read_target();
			polling_while_busy();
		}else{
			printf("Operation cancelled.\n");
			exit_bootmode();
			exit_app(1);
		}
	}
	if( arg_opt & ARG_EFUSE ) {
		printf("Change EFUSE to 0x%X. OK? [y/N] > ", fuse_val[FUSE_E]);
		char ans = getchar();
		if( ans == 'y' ) {
			write_efuse(fuse_val[FUSE_E]);
			read_target();
			polling_while_busy();
		}else{
			printf("Operation cancelled.\n");
			exit_bootmode();
			exit_app(1);
		}
	}

	if( arg_opt & ARG_ERASE ) {
		printf("Erasing Chip...\n");
		write_target(ERASE_CHIP);
		read_target();
		callback = expandFromArray(data);
		if( (callback & 0xFFFF00) == (0xAC80 << 8) ) {
			printf("    Erased flash\n");
			polling_while_busy();
		}else{
			printf("Failed to erase\n.");
			exit_app(1);
		}
	}

	if( arg_opt & ARG_FLASH ) {
		// HEXファイルをオープン・パース
		printf("Beginning flashing...\n");
		if( parse() ) {
			printf("    Parsing %s is OK!\n", filename);
			flash();
			printf("    Flash operation finished.\nThank you.\n");
		}else{
			printf("Failed to parse hex file %s.\n", filename);
			exit_app(1);
		}
		polling_while_busy();
	}

	exit_bootmode();
	FT_Close(handle);
	return 0;
}
